"""CLI script to dump pickle/data files"""
import sys
import pickle
from pprint import pprint

def main():
    """Read pickle file and pretty print it"""
    with open(sys.argv[1], "rb") as input_fd:
        pprint(pickle.load(input_fd))

if __name__ == '__main__':
    main()



"""Main entry point to load configuration and classes"""
import os
import time
import configparser
import importlib
from .util import configure

os.environ['TZ'] = 'UTC'
time.tzset()

class UnknownTraderModuleError(Exception):
    """Unknown trader module"""

class Botic:
    """Botic base class to setup and start the trader"""
    # pylint: disable=too-few-public-methods
    # pylint: disable=no-member
    def __init__(self, config_path: str, do_print=True) -> None:
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        configure(self, do_print=do_print)
        self._setup_trader()

    def _setup_trader(self) -> None:
        """Load the trader module specified in the config.

        Exchange config must follow this format for auto-import:
            config:
                [trader]
                trader_module = Simple
            code:
                trader/simple.py -> class Simple(BaseTrader): ...
        """
        mod_path = 'botic.trader.{}'.format(self.trader_module.lower())
        mod = importlib.import_module(mod_path) #, package='botic')
        obj = getattr(mod, self.trader_module, None)
        if not obj:
            raise UnknownTraderModuleError('Unknown trader module: {}'.format(self.trader_module))
        self.trader = obj(self.config)
        # Write config vars to trader object
        configure(self.trader, do_print=False)

    def run(self) -> None:
        """Entry point to start the bot"""
        self.trader.run()



"""Simple helper methods to be used throughout the code"""
import typing as t
from botic.defaults import CONFIG_DEFAULTS

def parse_datetime(value: str) -> str:
    """Parse datetime string with optional milliseconds and/or timezone

    Works for Coinbase but needs to be improved.

    Args:
        value (str): String representation of a date and time

    Returns:
        str: A new string without milliseconds and timezone
    """
    return str(value).split('.')[0].split('Z')[0]


def str2bool(value: str) -> bool:
    """Convert configuration string value to bool, if valid

    Args:
        value (str): String representation of a boolean value (e.g. yes, true, t, 1)

    Returns:
        bool: True or False
    """
    if isinstance(value, bool):
        return value
    return value.lower() in ("yes", "true", "t", "1")

def configure(obj, do_print=False) -> None:
    """Calls setattr() on configuration key,val pairs from obj.config, or sets defaults from
    CONFIG_DEFAULT.

    These are setup so they can be accessed like: obj.key
    It also sets key,value pairs from the [trader] section for the trader module to convert
    to its needs.

    Args:
        do_print (bool): If True, print and log non-auth configuration items
    """
    for section,items in CONFIG_DEFAULTS.items():
        for key, cast, default in items:
            val = getconf(obj.config, section, key, cast, default)
            # Special conversion to handle percent values and lists
            if key == 'mail_to':
                val = val.split(',')
            if not (section == 'exchange' and key in ('key', 'passphrase', 'b64secret')):
                if do_print:
                    print('config: [{}][{}] -> {}'.format(section, key, val))
            setattr(obj, key, val)
    # Now configure [trader] section, ingest all values as strings (trader module is responsible
    # for conversion)
    for key, val in obj.config['trader'].items():
        setattr(obj, key, val)

def getconf(config, section: str, key: str, cast: t.Type, default: t.Any) -> t.Any:
    """Converts configuration values to Python types

    Args:
        section (str): The configuration section of config (e.g. 'auth')
        key (str): The section's key to get()
        cast (type): Cast the retrieved value of config[section][key]

    Returns:
        any: The value of the configuration section->key
    """
    val = config[section].get(key, default)
    if cast == bool:
        val = str2bool(val)
    else:
        val = cast(val)
    return val



"""Test1 trader

Description:
    - Watch for price drops and buy the drop.
    - Sell at N% increase from current price.
"""
import time
from decimal import Decimal
import typing as t
import datetime
from .base import BaseTrader
from ..util import str2bool, parse_datetime
from ..exchange.exceptions import ExchangeSellLimitError

class Test1(BaseTrader):
    """Test1 trader"""
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=no-member
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-function-args
    def __init__(self, config) -> None:
        super().__init__(config)
        self.usd_decimal_places = None
        self.size_decimal_places = None
        self.current_price = None
        self.current_price_target = None
        self.taker_fee = None
        self.maker_fee = None
        self.usd_volume = None
        self.product_info = None
        self.current_price_increase = None
        self.wallet = None
        self.can_buy = False
        self.pdiff = 0.0
        self._rate_limit_log = time.time()
        self._tracker_prices = []
        self._tracker_times = []
        self._hodl_value = None
        self._neutral = None

    def configure(self) -> None:
        self.sell_target = Decimal(self.sell_target)/100
        self.buy_percent = Decimal(self.buy_percent)/100
        self.buy_max = Decimal(self.buy_max)
        self.buy_min = Decimal(self.buy_min)
        self.buy_drop = Decimal(self.buy_drop)/100
        self.tracker_time = int(self.tracker_time)

    def _time2datetime(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.exchange.get_time())

    def _append_tracker(self, price, timestamp):
        self._tracker_prices.append(price)
        self._tracker_times.append(timestamp)
        while 1:
            if timestamp - self._tracker_times[0] > self.tracker_time:
                del self._tracker_times[0]
                del self._tracker_prices[0]
            else:
                break

    def run_trading_algorithm(self) -> None:
        self.product_info = self.exchange.get_product_info()
        self.maker_fee, self.taker_fee, self.usd_volume = self.exchange.get_fees()
        self.size_decimal_places, self.usd_decimal_places = self.exchange.get_precisions()

        self.current_price = self.exchange.get_price()
        self.current_time = self.exchange.get_time()
        self._append_tracker(self.current_price, self.current_time)

        self._get_current_price_target()
        self.wallet = self.exchange.get_usd_wallet()

        if not self._hodl_value:
            self._hodl_value = self.wallet / self.current_price
            self._neutral = self.current_price

        self.can_buy = self._check_if_can_buy()
        self._maybe_buy_sell()
        self._check_sell_orders()
        hold_value = self.exchange.get_hold_value()
        if time.time() - self._rate_limit_log > 0.5:
            self._rate_limit_log = time.time()
            total_value = hold_value + self.wallet
            self.logit(
                'wallet:{:2f} open:{} price:{} held:{} canbuy:{} total-value:{} pdiff:{} hodl-value:{}'.format(
                self.wallet, self._total_open_orders, self.current_price, hold_value,
                self.can_buy, total_value, self.pdiff,
                round(self._hodl_value * self.current_price)),
                custom_datetime=self._time2datetime())

    def _get_current_price_target(self) -> Decimal:
        current_percent_increase = (self.maker_fee + self.taker_fee) + (self.sell_target)
        self.current_price_target = round(
            self.current_price * current_percent_increase + self.current_price,
            self.usd_decimal_places
        )
        self.current_price_increase = self.current_price * current_percent_increase
        return self.current_price_target

    @property
    def _total_open_orders(self) -> int:
        total = 0
        for _, order in self.data.items():
            if not order['completed'] and order['sell_order']:
                total += 1
        return total

    def _check_if_can_buy(self) -> bool:
        can = True
        diff = (self._neutral - self.current_price)/self._neutral
        if abs(diff) >= 0.05:
            can = True
        else:
            return False
        """
        self.pdiff = round(
            (self._tracker_prices[0]-self._tracker_prices[-1]) / self._tracker_prices[0], 5)
        if self.pdiff <= self.buy_drop:
            can = True
        else:
            can = False
            self.pdiff = round(
                (max(self._tracker_prices)-min(self._tracker_prices)) / max(self._tracker_prices), 5)
            if self.pdiff <= self.buy_drop:
                self.logit('_check_if_can_buy: condition 2 hit (max()) {}'.format(pdiff))
                can = True
        """
        # Now check if there are open sells near this position
        for _, order in self.data.items():  # self.open_sells:
            if order['completed']:
                continue
            sell_order = order['sell_order']
            if not sell_order:
                continue
            if not 'price' in sell_order:
                continue
            sell_price = Decimal(sell_order['price'])
            fees = self.maker_fee + self.taker_fee
            barrier = self.buy_barrier
            adjusted_sell_price = round(
                sell_price - ((Decimal(barrier) + fees) * sell_price),
                self.usd_decimal_places
            )
            if adjusted_sell_price <= self.current_price_target:
                can = False
                break
        if can:
            self._neutral = self.current_price
        return can

    def _maybe_buy_sell(self) -> None:
        assert self.wallet is not None, 'Wallet must be set.'
        assert self.current_price is not None, 'Current price must be set.'
        if not self.can_buy:
            return

        # Check if USD wallet has enough available
        if self.wallet < Decimal(self.product_info.min_market_funds):
            self.logit('WARNING: Wallet value too small (<${}): {}'.format(
                self.product_info.min_market_funds, self.wallet),
                custom_datetime=self._time2datetime()
            )
            return

        # Calculate & check if size is big enough (sometimes its not if wallet is too small)
        buy_amount = round(
            Decimal(self.buy_percent) * Decimal(self.wallet), self.usd_decimal_places
        )
        buy_size = round(Decimal(buy_amount) / self.current_price, self.size_decimal_places)
        if buy_size <= self.product_info.base_min_size:
            self.logit('WARNING: Buy size is too small {} {} < {} wallet:{}.'.format(
                buy_amount, buy_size, self.product_info.base_min_size, self.wallet),
                custom_datetime=self._time2datetime()
            )
            self.logit('DEBUG: {}'.format(self.product_info.config),
                custom_datetime=self._time2datetime())
            buy_amount = self.buy_min
            buy_size = round(Decimal(buy_amount) / self.current_price, self.size_decimal_places)
            self.logit('DEFAULT_BUY_SIZE_TO_MIN: {} {}'.format(buy_amount, buy_size),
                custom_datetime=self._time2datetime())

        # Check if USD wallet has enough available
        if buy_amount < Decimal(self.product_info.min_market_funds):
            self.logit('WARNING: Buy amount too small (<${}): {}'.format(
                self.product_info.min_market_funds, buy_amount),
                custom_datetime=self._time2datetime()
            )
            buy_amount = self.buy_min
            buy_size = round(Decimal(buy_amount) / self.current_price, self.size_decimal_places)
            self.logit('DEFAULT_BUY_SIZE_TO_MIN: {} {}'.format(buy_amount, buy_size),
                custom_datetime=self._time2datetime())

        # Make sure buy_amount is within buy_min/max
        if buy_amount < self.buy_min:
            self.logit('WARNING: buy_min hit. Setting to min.',
                custom_datetime=self._time2datetime())
            buy_amount = self.buy_min
        elif buy_amount > self.buy_max:
            self.logit('WARNING: buy_max hit. Setting to max.',
                custom_datetime=self._time2datetime())
            buy_amount = self.buy_max

        if Decimal(self.wallet) < Decimal(self.buy_min):
            self.logit('INSUFFICIENT_FUNDS', custom_datetime=self._time2datetime())
            return

        # adjust size to fit with fee
        buy_size = round(
            Decimal(buy_size) - Decimal(buy_size) * Decimal(self.taker_fee),
            self.size_decimal_places
        )
        self.logit('BUY: price:{} amount:{} size:{}'.format(
            self.current_price, buy_amount, buy_size),
            custom_datetime=self._time2datetime()
        )
        response = self.exchange.buy_market(buy_amount)
        self.logit('BUY-RESPONSE: {}'.format(response), custom_datetime=self._time2datetime())
        if 'message' in response:
            self.logit('WARNING: Failed to buy', custom_datetime=self._time2datetime())
            return
        order_id = response['id']
        errors = 0
        self.last_buy = None
        # Wait until order is completely filled
        if order_id in self.data:
            self.logit('ERROR: order_id exists in data. ????: {}'.format(order_id),
                custom_datetime=self._time2datetime())
        self.data[order_id] = {
            'first_status': response, 'last_status': None, 'time': self.exchange.get_time(),
            'sell_order': None, 'sell_order_completed': None,
            'completed': False, 'profit_usd': None
        }
        self.write_data()
        done = False
        status_errors = 0
        buy = {}
        while 1:
            try:
                buy = self.exchange.get_order(order_id)
                self.data[order_id]['last_status'] = buy
                self.write_data()
                if 'settled' in buy:
                    if buy['settled']:
                        self.logit('FILLED: size:{} funds:{}'.format(
                            buy['filled_size'], buy['funds']),
                            custom_datetime=self._time2datetime())
                        self.last_buy = buy
                        done = True
                        break
                else:
                    self._handle_failed_order_status(order_id, buy, status_errors)
                    status_errors += 1
                if status_errors > 10:
                    errors += 1
            except Exception as err:
                self.logit('WARNING: get_order() failed:' + str(err),
                    custom_datetime=self._time2datetime())
                errors += 1
                time.sleep(10)
            if errors > 5:
                self.logit('WARNING: Failed to get order. Manual intervention needed.: {}'.format(
                    order_id),
                    custom_datetime=self._time2datetime())
                break
            time.sleep(2)

        # Buy order done, now place sell
        if done:
            msg = 'BUY-FILLED: size:{} funds:{}\n'.format(buy['filled_size'], buy['funds'])
            self.logit(msg, custom_datetime=self._time2datetime())
            try:
                response = self.exchange.sell_limit(
                    self.current_price_target,
                    self.last_buy['filled_size']
                )
                self.logit('SELL-RESPONSE: {}'.format(response),
                    custom_datetime=self._time2datetime())
                msg = '{} SELL-PLACED: size:{} price:{}'.format(
                    msg, self.last_buy['filled_size'], self.current_price_target)
                for i in msg.split('\n'):
                    self.logit(i.strip(), custom_datetime=self._time2datetime())
                if not self.notify_only_sold:
                    self.send_email('BUY/SELL', msg=msg)
                self.data[order_id]['sell_order'] = response
            except ExchangeSellLimitError as err:
                self.logit('ExchangeSellLimitError: {}'.format(err),
                    custom_datetime=self._time2datetime())
                self.data[order_id]['completed'] = True
                self.data[order_id]['sell_order'] = None
            self.write_data()
            self.last_buy = None
        else:
            # buy was placed but could not get order status
            if 'message' in buy:
                msg = 'BUY-PLACED-NOSTATUS: {}\n'.format(buy['message'])
            else:
                msg = 'BUY-PLACED-NOSTATUS: size:{} funds:{}\n'.format(
                    buy['filled_size'], buy['funds'])
            self.logit(msg, custom_datetime=self._time2datetime())
            self.send_email('BUY-ERROR', msg=msg)

    def _handle_failed_order_status(self, order_id: str, status: t.Mapping[str, t.Any]) -> None:
        if 'message' in status:
            self.logit('WARNING: Failed to get order status: {}'.format(status['message']),
                custom_datetime=self._time2datetime())
            self.logit(
                'WARNING: Order status error may be temporary, due to coinbase issues or exchange '
                'delays. Check: https://status.pro.coinbase.com',
                custom_datetime=self._time2datetime()
            )
        else:
            self.logit('WARNING: Failed to get order status: {}'.format(order_id),
                custom_datetime=self._time2datetime())
        time.sleep(10)

    def _check_sell_orders(self) -> None:
        """ Check if any sell orders have completed """
        # pylint: disable=too-many-locals
        # pylint: disable=bare-except
        for buy_order_id, info in self.data.items():
            if self.data[buy_order_id]['completed']:
                continue
            if not info['sell_order']:
                self.logit('WARNING: No sell_order for buy {}. This should not happen.'.format(
                    buy_order_id), custom_datetime=self._time2datetime())
                if self.exchange.get_time() - info['time'] > 60 * 60 * 2:
                    self.logit('WARNING: Failed to get order status:',
                        custom_datetime=self._time2datetime())
                    self.logit('WARNING: Writing as done/error since it has been > 2 hours.',
                        custom_datetime=self._time2datetime())
                    self.data[buy_order_id]['completed'] = True
                    self.write_data()
                continue
            if 'message' in info['sell_order']:
                self.logit(
                    'WARNING: Corrupted sell order, mark as done: {}'.format(info['sell_order']),
                    custom_datetime=self._time2datetime())
                self.data[buy_order_id]['completed'] = True
                self.data[buy_order_id]['sell_order'] = None
                self.write_data()
                self.send_email('SELL-CORRUPTED',
                    msg='WARNING: Corrupted sell order, mark as done: {}'.format(
                        info['sell_order'])
                )
                time.sleep(3600)
                continue
            sell = self.exchange.get_order(info['sell_order']['id'])
            if 'message' in sell:
                self.logit('WARNING: Failed to get sell order status (retrying later): {}'.format(
                    sell['message']), custom_datetime=self._time2datetime())
                if self.exchange.get_time() - info['time'] > 60 * 60 * 2:
                    self.logit('WARNING: Failed to get order status:',
                        custom_datetime=self._time2datetime())
                    self.logit('WARNING: Writing as done/error since it has been > 2 hours.',
                        custom_datetime=self._time2datetime())
                    self.data[buy_order_id]['completed'] = True
                    self.write_data()
                continue

            if 'status' in sell and sell['status'] != 'open':
                # calculate profit from buy to sell
                # done, remove buy/sell
                self.data[buy_order_id]['completed'] = True
                self.data[buy_order_id]['sell_order_completed'] = sell
                if sell['status'] == 'done':
                    try:
                        first_time = self.data[buy_order_id]['first_status']['created_at']
                    except:
                        first_time = None
                    sell_value = Decimal(sell['executed_value'])
                    #sell_filled_size = Decimal(sell['filled_size'])
                    #buy_filled_size = Decimal(info['last_status']['filled_size'])
                    buy_value = Decimal(info['last_status']['executed_value'])
                    buy_sell_diff = round(sell_value - buy_value, 2)
                    if first_time:
                        done_at = time.mktime(
                            time.strptime(parse_datetime(first_time), '%Y-%m-%dT%H:%M:%S'))
                    else:
                        done_at = time.mktime(
                            time.strptime(parse_datetime(sell['done_at']), '%Y-%m-%dT%H:%M:%S'))
                    self.data[buy_order_id]['profit_usd'] = buy_sell_diff
                    msg = 'SOLD: duration:{:.2f} bought:{} sold:{} profit:{}'.format(
                        self.exchange.get_time() - done_at,
                        round(buy_value, 2),
                        round(sell_value, 2),
                        buy_sell_diff
                    )
                    self.logit(msg, custom_datetime=self._time2datetime())
                    self.send_email('SOLD', msg=msg)
                else:
                    self.logit('SOLD-WITH-OTHER-STATUS: {}'.format(sell['status']),
                        custom_datetime=self._time2datetime())
                self.write_data()



"""Base class of abstract methods that are to be implented by traders"""
import os
import time
import importlib
from random import uniform
from abc import abstractmethod
from ..basebot import BaseBot
from ..botic import configure

class UnknownExchangeModuleError(Exception):
    """Unknown trader module"""

class BaseTrader(BaseBot):
    """Base class of abstract methods that are to be implented by traders"""
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=no-member
    # pylint: disable=attribute-defined-outside-init
    def _init(self) -> None:
        """Initialize configuration, lock, data and load exchange"""
        self.configure()
        self.init_lock()
        self.init_data()
        self._load_exchange()

    def _load_exchange(self) -> None:
        """Load the exchange module specified in the config.

        Exchange config must follow this format for auto-import:
            config:
                [exchange]
                exchange_module = CoinbasePro
            code:
                exchange/coinbasepro.py -> class CoinbasePro(ExchangeBase): ...
        """
        mod_path = 'botic.exchange.{}'.format(self.exchange_module.lower())
        mod = importlib.import_module(mod_path)
        obj = getattr(mod, self.exchange_module, None)
        if not obj:
            raise UnknownExchangeModuleError('Unknown exchange module: {}'.format(
                self.exchange_module))
        self.exchange = obj(self.config)
        self.exchange.authenticate()
        configure(self.exchange, do_print=False)

    @abstractmethod
    def configure(self) -> None:
        """Method to convert key,value pairs set from the [trader] config section."""

    @abstractmethod
    def run_trading_algorithm(self) -> None:
        """Run the traders main algorithim. This is responsible for interacting with the exchange
        to fetch data and place buys/sells.
        """

    def run(self) -> None:
        """Main program loop"""
        self._init()
        # Throttle startups randomly
        time.sleep(uniform(1, 5))
        while 1:
            if os.path.exists(self.pause_file):
                self.logit('PAUSE')
                time.sleep(30)
                continue
            self.run_trading_algorithm()
            time.sleep(self.sleep_seconds)



"""Backtest exchange module

- Currently handles BTC-USD for the ProductInfo config.
- TODO: Add more

"""
import sys
import time
import configparser
import typing as t
import gzip
import datetime
import uuid
import decimal
from pkgutil import get_data
from .base import BaseExchange, ProductInfo, Decimal

class Backtest(BaseExchange):
    """Backtest"""
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=no-member
    def __init__(self, config: configparser.ConfigParser) -> None:
        super().__init__(config)
        self.usd_decimal_places = 2
        self.size_decimal_places = 8
        self._last_call = time.time()
        self._wallet = Decimal('10000.00')
        self._orders = {}
        self._last_price = None
        self._maker_fee = Decimal('0.0010')
        self._taker_fee = Decimal('0.0020')
        self._adjusted_time = 0.0

        # TODO: Track volume and adjust fees according to;
        # https://help.coinbase.com/en/pro/trading-and-funding/trading-rules-and-fees/fees
        self._product_info_config = {
            'id':'BTC-USD',
            'display_name':'BTC/USD',
            'base_currency':'BTC',
            'quote_currency':'USD',
            'base_increment':Decimal('0.00000001'),
            'quote_increment':Decimal('0.01000000'),
            'base_min_size':Decimal('0.00100000'),
            'base_max_size':Decimal('280.00000000'),
            'min_market_funds':30,
            'max_market_funds':1000000,
            'status':'online',
            'status_message':'',
            'cancel_only':False,
            'limit_only':False,
            'post_only':False,
            'trading_disabled':False,
            'margin_enabled':False,
        }
        #if self.coin != 'BTC-USD':
        #    raise Exception('Currently only handles BTC-USD')
        gdata = get_data('botic', 'data/historical-btc.csv.gz')
        self._data = gzip.decompress(gdata).decode("utf-8").split('\n')
        self._data_pos = 1
        self._data_buf = []
        self._data_buf_idx = 0
        #"timestamp","low","high","open","close","volume"

    def _prepare_candle(self):
        """
        Convert data frame/candle to a list with filled in points between:
            open -> low/high -> close
        Currently tries something like:
            open -> (open+low)/2 -> (open+low+high)/3 -> high -> (close+low+high)/3 -> close
        """
        try:
            data = self._data[self._data_pos].replace('"', '').split(',')
            tstamp, low, high, x_open, x_close, volume  = [Decimal(i) for i in data]
        except decimal.InvalidOperation:
            self.logit('Backtest has ended. No more data.')
            sys.exit(0)

        if float(tstamp) < self._adjusted_time:
            raise Exception('Time went backwards: tstamp:{} was:{}'.format(
                tstamp, self._adjusted_time))

        self._adjusted_time = float(tstamp)
        # for a check of date
        self._time2datetime()
        #self.adjusted_time = datetime.datetime.fromtimestamp(tstamp)
        self._data_pos += 1
        self._data_buf = []
        open_low = round((x_open+low)/2, 2)
        open_low_high = round((x_open+low+high)/3, 2)
        close_low_high = round((x_close+low+high)/3, 2)
        self._data_buf.append(x_open)
        if not open_low in self._data_buf:
            self._data_buf.append(open_low)
        if not open_low_high in self._data_buf:
            self._data_buf.append(open_low_high)
        if not high in self._data_buf:
            self._data_buf.append(high)
        if not close_low_high in self._data_buf:
            self._data_buf.append(close_low_high)
        if not x_close in self._data_buf:
            self._data_buf.append(x_close)
        self._data_buf_idx = 0

    def authenticate(self):
        return None

    def get_price(self) -> Decimal:
        if not self._data_buf:
            self._prepare_candle()
        price = self._data_buf[self._data_buf_idx]
        self._data_buf_idx += 1
        if self._data_buf_idx >= len(self._data_buf):
            self._prepare_candle()
        self._last_price = Decimal(price)
        self._settle_trades()
        #self._adjusted_time += 1
        return self._last_price

    def _settle_trades(self):
        """Settle trades
        TODO: settle limit buys
        """
        for order_id, val in self._orders.items():
            if val['side'] == 'sell' and val['status'] == 'open':
                price = Decimal(val['price'])
                size = Decimal(val['size'])
                # Sell was filled if latest price is >=
                if price <= self._last_price:
                    val['status'] = 'done'
                    val['done_at'] = self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                    val['done_reason'] = 'filled'
                    usd_used = size*price
                    fees = round(usd_used * self._maker_fee, 12)
                    executed_value = usd_used - fees
                    self._wallet += executed_value
                    val['executed_value'] = str(executed_value)
                    val['fill_fees'] = str(fees)
                    val['settled'] = True
                    print('exchange-settled: {} last_price:{} -> order:{} / {}'.format(
                        val['id'], self._last_price, price, size))

    def get_precisions(self) -> ProductInfo:
        product_info = self.get_product_info()
        base_increment = '%.12f' % (product_info.base_increment)
        quote_increment = '%.12f' % (product_info.quote_increment)
        self.size_decimal_places = base_increment.split('1')[0].count('0')
        self.usd_decimal_places = quote_increment.split('1')[0].count('0')
        return (self.size_decimal_places, self.usd_decimal_places)

    def get_product_info(self) -> ProductInfo:
        product_info = ProductInfo(self._product_info_config)
        return product_info

    def get_usd_wallet(self) -> Decimal:
        return self._wallet

    def get_open_sells(self) -> t.List[t.Mapping[str, Decimal]]:
        sells = []
        for _, val in self._orders.items():
            if val['side'] == 'sell' and val['status'] == 'open':
                val['price'] = Decimal(val['price'])
                val['size'] = Decimal(val['size'])
                sells.append(val)
        return sells

    def get_fees(self) -> t.Tuple[Decimal, Decimal, Decimal]:
        return (self._maker_fee, self._taker_fee, Decimal('1'))

    def buy_limit(self, price: Decimal, size: Decimal) -> dict:
        #fixed_price = str(round(Decimal(price), self.usd_decimal_places))
        #fixed_size = str(round(Decimal(size), self.size_decimal_places))
        raise Exception('Not implemented: buy_limit')
        return {}

    def buy_market(self, funds: Decimal) -> dict:
        if funds > self._wallet:
            return {'message':'Insufficient funds'}
        self._wallet = self._wallet - funds
        funds = round(Decimal(funds), self.usd_decimal_places)
        # funds - fees
        fees = round(funds * self._taker_fee, 12)
        executed_value = round(funds - (fees), 12)
        filled_size = round(executed_value/self._last_price, 8)
        uid = str(uuid.uuid4())
        response = {
            'created_at': self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            'done_at': self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            'done_reason': 'filled',
            'executed_value': str(executed_value),
            'fill_fees': str(fees),
            'filled_size': str(filled_size),
            'funds': str(executed_value),
            'id': uid,
            'post_only': False,
            'product_id': 'BTC-USD',
            'profile_id': 'xyz',
            'settled': True,
            'side': 'buy',
            'specified_funds': str(funds),
            'status': 'done',
            'type': 'market'
        }
        self.logit('buy_market: funds:{}'.format(funds), custom_datetime=self._time2datetime())
        self._orders[uid] = response
        return response

    def sell_limit(self, price: Decimal, size: Decimal) -> dict:
        fixed_price = str(round(Decimal(price), self.usd_decimal_places))
        fixed_size = str(round(Decimal(size), self.size_decimal_places))
        self.logit('sell_limit: price:{} size:{}'.format(fixed_price, fixed_size),
            custom_datetime=self._time2datetime())
        uid = str(uuid.uuid4())
        response = {
            'created_at': self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            'executed_value': '0',
            'fill_fees': '0',
            'filled_size': '0',
            'id': uid,
            'post_only': False,
            'price': fixed_price,
            'product_id': 'BTC-USD',
            'settled': False,
            'side': 'sell',
            'size': fixed_size,
            'status': 'open',
            'stp': 'dc',
            'time_in_force': 'GTC',
            'type': 'limit'
        }
        self._orders[uid] = response
        return response

    def sell_market(self, size: Decimal) -> dict:
        fixed_size = str(round(Decimal(size), self.size_decimal_places))
        self.logit('sell_market: size:{}'.format(fixed_size),
            custom_datetime=self._time2datetime())
        uid = str(uuid.uuid4())
        usd_used = Decimal(size) * self._last_price
        fees = round(usd_used * self._taker_fee, 12)
        executed_value = usd_used - fees
        self._wallet += executed_value
        response = {
            'created_at': self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            'done_at': self._time2datetime().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            'executed_value': str(executed_value),
            'fill_fees': str(fees),
            'filled_size': str(size),
            'id': uid,
            'post_only': False,
            'price': str(self._last_price),
            'product_id': 'BTC-USD',
            'settled': True,
            'side': 'sell',
            'size': fixed_size,
            'status': 'done',
            'stp': 'dc',
            'time_in_force': 'GTC',
            'type': 'market'
        }
        self._orders[uid] = response
        return response

    def cancel(self, order_id: str) -> bool:
        if self._orders[order_id]['status'] == 'open':
            self._orders[order_id]['status'] = 'cancel'
            self._orders[order_id]['settled'] = True
        return self._orders[order_id]

    def get_order(self, order_id: str) -> dict:
        return self._orders[order_id]

    def _time2datetime(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self._adjusted_time)

    def get_time(self) -> float:
        return self._adjusted_time

    def get_hold_value(self) -> Decimal:
        total = Decimal('0.0')
        price = self._last_price
        for _, info in self._orders.items():
            if info['side'] == 'sell' and info['status'] == 'open':
                amount = Decimal(info['size']) * Decimal(self._last_price)
                total += amount
        return total



"""CLI entry points for Botic
"""
import os
import sys
import time
from shlex import quote
from .botic import Botic

os.environ['TZ'] = 'UTC'
time.tzset()

def usage():
    """Prints the usage information"""
    print('{} <config-file>'.format(sys.argv[0]))
    sys.exit(1)

def main_persist():
    """Run botic in a infinite loop"""
    if len(sys.argv) != 2:
        usage()
    try:
        while 1:
            os.system('botic {}'.format(quote(sys.argv[1])))
            time.sleep(60)

    except KeyboardInterrupt:
        print('Exiting...')

def main() -> None:
    """Run botic"""
    if len(sys.argv) != 2:
        usage()
    config_path = sys.argv[1]
    bot = Botic(config_path)
    try:
        bot.run()
    except KeyboardInterrupt:
        print('exit')


if __name__ == '__main__':
    main()



"""Base class for handling standard things like data files and logs
"""
import os
import sys
import time
import pickle
import smtplib
import typing as t
from datetime import datetime
from abc import ABCMeta
from filelock import FileLock

os.environ['TZ'] = 'UTC'
time.tzset()

class DataNameError(Exception):
    """Invalid data name"""

class BaseBot(metaclass=ABCMeta):
    """The base class for exchanges and trading bots

    See CONFIG_DEFAULTS for more information on attributes that are set during _configure().

    Args:
        config_path (str): The path to the configuration file

    Attributes:
        config (configparser.ConfigParser): ConfigParse object from config_path
    """
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=too-few-public-methods
    # pylint: disable=no-self-use
    # pylint: disable=bare-except
    # pylint: disable=no-member
    def __init__(self, config) -> None:
        self.config = config


    def init_data(self) -> None:
        """Verify data_file name and load existing data if it exists"""
        self.data = {}
        if not self.data_file.endswith('.data'):
            raise DataNameError('ERROR: Data filenames must end in .data')
        if os.path.exists(self.data_file):
            with open(self.data_file, "rb") as data_fd:
                self.data = pickle.load(data_fd)

    def init_lock(self) -> None:
        """Initialize and open lockfile.

        Note that this is done to avoid running duplicate configurations at the same time.
        """
        self.lock_file = self.data_file.replace('.data', '.lock')
        self.lock = FileLock(self.lock_file, timeout=1)
        try:
            self.lock.acquire()
        except:
            print('ERROR: Failed to acquire lock: {}'.format(self.lock_file))
            print('Is another process already running with this config?')
            sys.exit(1)

    def write_data(self) -> None:
        """Write self.data to disk atomically"""
        with open(self.data_file + '-tmp', "wb") as data_fd:
            pickle.dump(self.data, data_fd)
            os.fsync(data_fd)
        if os.path.exists(self.data_file):
            os.rename(self.data_file, self.data_file + '-prev')
        os.rename(self.data_file + '-tmp', self.data_file)

    def _log(self, path: t.AnyStr, msg: t.Any, custom_datetime=None) -> None:
        """TODO: Replace me with Python logging"""
        if custom_datetime:
            now = custom_datetime
        else:
            now = datetime.now()
        print('{} {}'.format(now, str(msg).strip()))
        if not self.log_disabled:
            with open(path, 'a') as log_fd:
                log_fd.write('{} {}\n'.format(now, str(msg).strip()))

    def logit(self, msg: t.Any, custom_datetime=None) -> None:
        """TODO: Replace me with Python logging"""
        if not self.coin in msg:
            msg = '{} {}'.format(self.coin, msg)
        self._log(self.log_file, msg, custom_datetime=custom_datetime)

    def send_email(self, subject: str, msg: t.Optional[t.AnyStr] = None) -> None:
        """TODO: Add auth, currently setup to relay locally or relay-by-IP"""
        for email in self.mail_to:
            if not email.strip():
                continue
            headers = "From: %s\r\nTo: %s\r\nSubject: %s %s\r\n\r\n" % (
                self.mail_from, email, self.coin, subject)
            if not msg:
                msg2 = subject
            else:
                msg2 = msg
            msg2 = headers + msg2
            server = smtplib.SMTP(self.mail_host)
            server.sendmail(self.mail_from, email, msg2)
            server.quit()
            time.sleep(0.1)



"""Query CoinbasePro public API for historical data and save as CSV file"""
import sys
import time
import datetime
from datetime import timedelta
import cbpro

SIZE = 60

def date2str(dtobj_or_str):
    """Convert date to str"""
    return str(dtobj_or_str).replace(' ', 'T').split('.')[0]+'Z'

def generate_historical_csv(outfile, coin='BTC-USD', days_ago=93):
    """Generate CSV file from coinbase get_product_historic_rates"""
    # pylint: disable=too-many-locals
    public_client = cbpro.PublicClient()
    start_date = datetime.datetime.now() - timedelta(days=days_ago)
    end_date = datetime.datetime.now()
    next_date = start_date + timedelta(minutes=SIZE)
    out_fd = open(outfile, 'w')
    out_fd.write('"timestamp","low","high","open","close","volume"\n')
    while next_date < end_date:
        #print(next_date)
        stats = public_client.get_product_historic_rates(
            coin,
            granularity=SIZE,
            start=date2str(next_date),
            end=date2str(next_date+timedelta(hours=5)))
        if 'message' in stats:
            print(stats['message'])
            sys.exit(1)
        next_date = next_date + timedelta(hours=5)
        # stats are from newest to oldest, so reverse it
        stats.reverse()
        for i in stats:
            (tstamp, low, high, x_open, x_close, x_volume) = i
            #dt = datetime.datetime.fromtimestamp(tstamp)
            #print(dt, x_close)
            out_fd.write('"{}","{}","{}","{}","{}","{}"\n'.format(
                tstamp, low, high, x_open, x_close, x_volume))
        time.sleep(1.1)



"""Exchange specific errors"""

class PriceError(Exception):
    """Could not convert price string to Decimal"""

class ExchangeError(Exception):
    """Generic exchange error"""

class ExchangeGetOrdersError(Exception):
    """Exchange get orders failed"""

class ExchangeAuthError(Exception):
    """Exchange authentication failed"""

class ExchangeBuyLimitError(Exception):
    """Exchange limit buy failed"""

class ExchangeBuyMarketError(Exception):
    """Exchange market buy failed"""

class ExchangeSellLimitError(Exception):
    """Exchange limit sell failed"""

class ExchangeSellMarketError(Exception):
    """Exchange market sell failed"""

class ExchangeProductInfoError(Exception):
    """Exchange product info failed"""

class ExchangeCancelError(Exception):
    """Exchange cancel order failed"""

class ExchangeFeesError(Exception):
    """Exchange get fees failed"""

class ExchangeWalletError(Exception):
    """Exchange get accounts failed"""



import os
import re
import sys
import glob
import pickle
import time
from collections import OrderedDict
from datetime import datetime
from decimal import Decimal
from operator import getitem
import curses
from curses import endwin
import cbpro

os.environ['TZ'] = 'UTC'
time.tzset()
PRICE_CACHE = {}
PRICE_CACHE_RATE = 93.1331 #update every N seconds
REFRESH = 5.5 #in seconds

def pdiff(old, new):
    try:
        return round(( (Decimal(new) - Decimal(old)) / Decimal(old)) * Decimal('100.0'), 2)
    except:
        return 'unk'

def parse_datetime(d):
    return str(d).split('.')[0].split('Z')[0]

def get_current_price(coin):
    last_update = time.time()
    current_price = Decimal('0.0')
    if not coin in PRICE_CACHE:
        public_client = cbpro.PublicClient()
        ticker = public_client.get_product_ticker(product_id=coin)
        try:
            current_price = Decimal(ticker['price'])
        except Exception as err:
            return None
    else:
        # check cache age
        if time.time() - PRICE_CACHE[coin]['last_update'] > PRICE_CACHE_RATE:
            public_client = cbpro.PublicClient()
            ticker = public_client.get_product_ticker(product_id=coin)
            current_price = Decimal(ticker['price'])
        else:
            last_update = PRICE_CACHE[coin]['last_update']
            current_price  = PRICE_CACHE[coin]['price']
    PRICE_CACHE[coin] = {'price':current_price, 'last_update':last_update}
    return Decimal(current_price)

def sec2time(sec):
    """ TODO: Replace this function as it is buggy """
    if hasattr(sec,'__len__'):
        return [sec2time(s) for s in sec]
    m, s = divmod(sec, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    pattern = r'%2dh %2dm %2ds'
    if h == 0 and m == 0 and d == 0:
        return r'%2ds' % (s)
    if h == 0 and d == 0:
        return r'%2dm %2ds' % (m, s)
    if d == 0:
        #return pattern % (h, m, s)
        return r'%2dh %2dm %2ds'  % (h, m, s)
    return ('%dd' + pattern) % (d, h, m, s)


def avg(l):
    if len(l) < 1:
        return Decimal(0.0)
    return Decimal(sum(l))/len(l)


def get_open_orders(regex_str):
    files = glob.glob(sys.argv[1]+'/*.data')
    stats = {}
    stats_incomplete = {}
    recent = []
    cur_time = time.time()
    open_times = []
    profit_dates = {}
    output = []
    open_orders_title = '{:>15} {:>9} {:>13} {:>13} {:>14} {:>11}'.format(
        'Duration', 'Coin', 'Bought', 'Sell-Price', 'Size', 'Diff%',
    )
    output.append(open_orders_title)
    for f in files:
        data = None
        coin = None
        with open(f, "rb") as fd:
            data = pickle.load(fd)
        if not data:
            continue
        for order_id in data:
            v = data[order_id]
            data[order_id]['created_at'] = v['first_status']['created_at']
        sorted_data = OrderedDict(sorted(data.items(), key = lambda x: getitem(x[1], 'created_at'), reverse=True))
        for order_id, v in sorted_data.items():
            if not v['first_status']:
                continue
            coin = v['first_status']['product_id']
            if regex_str and not re.search(regex_str, coin, re.IGNORECASE):
                continue
            if v['completed'] or not v['sell_order']:
                continue
            sell = v['sell_order']
            created_at = time.mktime(time.strptime(parse_datetime(sell['created_at']), '%Y-%m-%dT%H:%M:%S'))
            duration = cur_time - created_at
            try:
                price = sell['price']
            except Exception as err:
                price = str(err)
            size = sell['size']
            bought_price = round(Decimal(v['last_status']['executed_value']) / Decimal(v['last_status']['filled_size']), 4)
            cur_price = get_current_price(sell['product_id'])
            output.append('{:>15} {:>9} {:>13} {:>13} {:>13} {:>11}'.format(
                sec2time(duration),
                sell['product_id'],
                bought_price, price,
                size,
                pdiff(bought_price, cur_price),
                #round(cur_price - bought_price, 2)
            ))
    return output

def get_stats(regex_str):
    files = glob.glob(sys.argv[1]+'/*.data')
    stats = {}
    stats_incomplete = {}
    recent = []
    cur_time = time.time()
    open_times = []
    profit_dates = {}
    open_percents = []
    output = []
    output_open_orders = []
    output_daily_profits = []
    output_recent = []
    for f in files:
        data = None
        coin = None
        with open(f, "rb") as fd:
            data = pickle.load(fd)
        if not data:
            continue

        for order_id, v in data.items():
            if not v['first_status']:
                continue
            coin = v['first_status']['product_id']
            if regex_str:
                if not re.search(regex_str, coin, re.IGNORECASE):
                    continue
            if not coin in stats:
                stats[coin] = {
                    'epoch_diffs':[], 'profits':[], 'profits_total':Decimal('0.0'),
                    'open_orders':0, 'done_orders':0, 'avg_close_time':0.0, 'error_orders':0,
                }
            first_status = v['first_status']
            epoch = time.mktime(time.strptime(parse_datetime(first_status['created_at']), '%Y-%m-%dT%H:%M:%S'))
            if v['completed'] and 'sell_order_completed' in v and v['sell_order_completed'] and v['profit_usd']:
                # NOTE: Getting some completed orders w/o all the information filled in (done_at)
                # This seems to happen when failing to retreive status in the bot code
                date_only = v['sell_order_completed']['done_at'].split('T')[0]
                if not date_only in profit_dates:
                    profit_dates[date_only] = []
                profit_dates[date_only].append(v['profit_usd'])
                end_epoch = time.mktime(time.strptime(parse_datetime(v['sell_order_completed']['done_at']), '%Y-%m-%dT%H:%M:%S'))
                epoch_diff = end_epoch - epoch
                cur_diff = cur_time - end_epoch
                if cur_diff < (86400/12):
                    recent.append((coin, v))
                profit = v['profit_usd']
                stats[coin]['epoch_diffs'].append(epoch_diff)
                stats[coin]['profits'].append(profit)
                stats[coin]['profits_total'] += profit
                stats[coin]['done_orders'] += 1
            elif v['completed']:
                stats[coin]['error_orders'] += 1
            else:
                cur_price = get_current_price(coin)
                try:
                    cur_perc = (100*(cur_price/Decimal(v['sell_order']['price']))) - Decimal('100.0')
                    open_percents.append(cur_perc)
                except Exception as err:
                    # I think sometimes the state drifts after a cancel
                    # and v['sell'] was removed but v['completed'] is not True yet
                    #print('ERR:', err, v['sell_order'])
                    pass
                start_epoch = time.mktime(time.strptime(parse_datetime(v['first_status']['created_at']), '%Y-%m-%dT%H:%M:%S'))
                open_times.append(cur_time - start_epoch)
                stats[coin]['open_orders'] += 1

    sorted_keys = OrderedDict(sorted(stats.items(), key = lambda x: getitem(x[1], 'profits_total'), reverse=True))
    output.append('{:>8} {:>13} {:>7} {:>7} {:>7} {:>12} {:>19}'.format(
        'Coin',
        'Profits',
        'Open',
        'Done',
        'Error',
        'Avg-Profit',
        'Avg-Time',
    ))
    total_profits = Decimal('0.0')
    total_open_orders = 0
    total_done_orders = 0
    total_error_orders = 0
    agg_epoch = []
    agg_profits = []
    for key,v  in sorted_keys.items():
        coin = key
        if regex_str and not re.search(regex_str, coin, re.IGNORECASE):
            continue
        output.append('{:>8} {:>13} {:>7} {:>7} {:>7} {:>12} {:>19}'.format(
            coin,
            '$'+str(round(sum(v['profits']), 2)),
            v['open_orders'],
            v['done_orders'],
            v['error_orders'],
            '$'+str(round(avg(stats[coin]['profits']), 2)),
            sec2time(round(avg(v['epoch_diffs']), 2)) if v['epoch_diffs'] else 'None',

        ))
        agg_epoch.append(round(avg(v['epoch_diffs']), 2) if v['epoch_diffs'] else Decimal('0.0'))
        agg_profits.append(round(avg(stats[coin]['profits']), 2))
        total_open_orders += v['open_orders']
        total_done_orders += v['done_orders']
        total_error_orders += v['error_orders']
        total_profits += round(sum(v['profits']), 2)
    output.append('{:>8} {:>13} {:>7} {:>7} {:>7} {:>12} {:>19}'.format(
            'all',
            '$'+str(total_profits),
            total_open_orders,
            total_done_orders,
            total_error_orders,
            '$'+str(round(avg(agg_profits), 2)),
            sec2time(round(avg(agg_epoch), 2)),

    ))
    if open_times:
        min_open_time = sec2time(round(min(open_times), 2))
        max_open_time = sec2time(round(max(open_times), 2))
        avg_open_time = sec2time(round(avg(open_times), 2))
    else:
        min_open_time = Decimal('0.0')
        max_open_time = Decimal('0.0')
        avg_open_time = Decimal('0.0')
    output_open_orders.append('{:>16} {:>16} {:>16} {:>16}'.format( 'Open order times', 'Min', 'Max', 'Avg'))
    output_open_orders.append('{:>16} {:>16} {:>16} {:>16}'.format(' ', min_open_time, max_open_time, avg_open_time))
    cur_drift = round(avg(open_percents), 2)
    if cur_drift < 0:
        output_open_orders.append('Avg-drift: {}%'.format(cur_drift))
    else:
        output_open_orders.append('Avg-drift: {}%'.format(cur_drift))

    # Last 7 days with profits
    sorted_dates_val = OrderedDict(sorted(profit_dates.items(), key = lambda x: x[1], reverse=True))
    sorted_dates = sorted(profit_dates.keys(), reverse=True)
    x = []
    y = []
    for key in sorted_dates[:7]:
        if regex_str and not re.search(regex_str, key, re.IGNORECASE):
            continue
        val = profit_dates[key]
        date_total = round(sum(val), 2)
        x.append(key)
        y.append(date_total)
    if y:
        total_profit = []
        max_y = max(y)
        width = 50
        for i in range(len(y)):
            key = x[i]
            yy = y[i]
            nstars = int((yy/max_y) * width)
            output_daily_profits.append('{:>11} {} {}'.format(key, '$'+str(yy), '*'*nstars))
            total_profit.append(yy)
        nstars = int((avg(total_profit)/max_y) * width)
        output_daily_profits.append('{:>11} {} {}^'.format('Daily-Avg ', '$'+str(round(avg(total_profit), 2)), ' '*(nstars-1)))
    if recent:
        output_recent.append('{}'.format( 'Recently completed orders', ))
        output_recent.append('    {:>8} {:>11} {:>17} {:>19}'.format('Coin', 'Profit', 'Duration', 'Completed', ))
        # bubble sort, why not
        for i in range(len(recent)-1):
            for j in range(len(recent)-i-1):
                if recent[j][1]['sell_order_completed']['done_at'] < recent[j+1][1]['sell_order_completed']['done_at']:
                    tmp = recent[j+1]
                    recent[j+1] = recent[j]
                    recent[j] = tmp

        for coin, v in recent:
            if regex_str and not re.search(regex_str, coin, re.IGNORECASE):
                continue
            first_status = v['first_status']
            epoch = time.mktime(time.strptime(parse_datetime(first_status['created_at']), '%Y-%m-%dT%H:%M:%S'))
            end_epoch = time.mktime(time.strptime(parse_datetime(v['sell_order_completed']['done_at']), '%Y-%m-%dT%H:%M:%S'))
            epoch_diff = end_epoch - epoch
            cur_diff = cur_time - end_epoch
            profit = round(v['profit_usd'], 2)
            output_recent.append('    {:>8} {:>11} {:>17} {:>19}'.format(
                coin, '$'+str(profit), sec2time(epoch_diff), str(sec2time(cur_diff))+' ago')
            )
    return (output, output_open_orders, output_daily_profits, output_recent)


def addstr_wrap(stdscr, a1, a2, a3=None):
    try:
        if a3:
            stdscr.addstr(a1, a2, a3)
        else:
            stdscr.addstr(a1, a2)
    except:
        pass

def draw_menu(stdscr):
    k = 0
    cursor_x = 0
    cursor_y = 0
    filter_str_title = 'Filter: '
    filter_str = None
    filter_regex = None
    regex_str = None
    filter_capture = False
    can_draw = False
    last_draw = time.time()
    last_open_orders = time.time()
    last_stats = time.time()
    open_orders = get_open_orders(regex_str)
    mode_str = 'Stats'
    open_orders_title = '{:>15} {:>9} {:>13} {:>13} {:>13} {:>11}'.format(
        'Duration', 'Coin', 'Bought', 'Sell-Price', 'Size', 'Diff%',
    )
    stats = get_stats(regex_str)
    curses.use_default_colors()
    # nonblocking so it can fetch stats on an interval
    stdscr.nodelay(True)
    stdscr.clear()
    stdscr.refresh()
    # 0:black, 1:red, 2:green, 3:yellow, 4:blue, 5:magenta, 6:cyan, and 7:white.
    # init_pair: number,text,background
    curses.start_color()
    curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_WHITE)
    curses.init_pair(2, curses.COLOR_CYAN, -1)
    curses.init_pair(3, curses.COLOR_RED, -1)
    curses.init_pair(4, curses.COLOR_GREEN, -1)
    curses.init_pair(5, curses.COLOR_YELLOW, -1)
    curses.init_pair(6, curses.COLOR_BLUE, -1)
    curses.init_pair(7, curses.COLOR_MAGENTA, -1)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_CYAN)

    while 1:
        cur_time = time.time()
        # Initialization
        if cur_time - last_draw > 0.5:
            can_draw = True
            last_draw = cur_time
            stdscr.clear()
        else:
            can_draw = False

        height, width = stdscr.getmaxyx()

        if k == curses.KEY_DOWN:
            cursor_y = cursor_y + 1
        elif k == curses.KEY_UP:
            cursor_y = cursor_y - 1
        elif k == curses.KEY_RIGHT:
            cursor_x = cursor_x + 1
        elif k == curses.KEY_LEFT:
            cursor_x = cursor_x - 1

        cursor_x = max(0, cursor_x)
        cursor_x = min(width-1, cursor_x)

        cursor_y = max(0, cursor_y)
        cursor_y = min(height-1, cursor_y)

        # Declaration of strings
        if mode_str == 'Stats':
            statusbar_str = '{} | [q]uit [f]ilter [o]rders | {}'.format(
                mode_str, parse_datetime(datetime.now()))
        elif mode_str == 'Orders':
            statusbar_str = '{} | [q]uit [f]ilter [s]tats | {}'.format(
                mode_str, parse_datetime(datetime.now()))
        # Capture Filter regex and compile on return press
        # if filter_capture render bottom input bar
        if filter_capture:
            if not filter_str:
                filter_str = filter_str_title
            else:
                if k not in (0, -1):
                    if k == 10:
                        regex_str = filter_str.replace(filter_str_title, '').strip()
                        if not regex_str:
                            filter_regex = None
                        else:
                            try:
                                filter_regex = re.compile(regex_str, flags=re.IGNORECASE)
                            except:
                                # Report error?
                                filter_regex = None
                                error_msg = 'ERROR: Failed to compile filter regex!'
                                stdscr.attron(curses.color_pair(1))
                                stdscr.attron(curses.A_BOLD)
                                addstr_wrap(stdscr, height-1, 0, error_msg)
                                addstr_wrap(stdscr, height-1, len(error_msg), " " * (width - len(error_msg) - 1))
                                stdscr.attroff(curses.color_pair(1))
                                stdscr.attroff(curses.A_BOLD)
                        filter_str = None
                        filter_capture = False
                    elif k == 263 and len(filter_str) > len(filter_str_title):
                        filter_str = filter_str[:-1]
                    elif k not in (263, 10):
                        filter_str += chr(k)
            if filter_capture:
                stdscr.attron(curses.color_pair(1))
                addstr_wrap(stdscr, height-1, 0, filter_str)
                addstr_wrap(stdscr, height-1, len(filter_str), " " * (width - len(filter_str) - 1))
                stdscr.attroff(curses.color_pair(1))

        # Render bottom filter bar if not capturing and set
        if not filter_capture and regex_str:
            stdscr.attron(curses.color_pair(1))
            addstr_wrap(stdscr, height-1, 0, regex_str)
            addstr_wrap(stdscr, height-1, len(regex_str), " " * (len(regex_str)))
            stdscr.attroff(curses.color_pair(1))

        # Render status bar
        stdscr.attron(curses.color_pair(8))
        stdscr.attron(curses.A_BOLD)
        addstr_wrap(stdscr, 0, 0, statusbar_str)
        addstr_wrap(stdscr, 0, len(statusbar_str), " " * (width - len(statusbar_str)))
        stdscr.attroff(curses.color_pair(8))
        stdscr.attroff(curses.A_BOLD)
        #stdscr.hline(1, 0, '-', width)

        # Open orders section
        if mode_str == 'Orders':
            if cur_time - last_open_orders > REFRESH:
                last_open_orders = cur_time
                open_orders = get_open_orders(regex_str)
            stdscr.attron(curses.color_pair(7))
            stdscr.attron(curses.A_BOLD)
            i = 0
            #open_orders = ['1','2','3','4','5','6','7','8','9','10','11']
            for i,s in enumerate(open_orders):
                if i+5 > height:
                    break
                addstr_wrap(stdscr, 2+i, 0, s)
                addstr_wrap(stdscr, 2+i, len(s), " " * (width - len(s)))
            if len(open_orders) > 1:
                ndisplayed = '({}/{} displayed)'.format(i, len(open_orders)-1)
                addstr_wrap(stdscr, 3+i, 0, ndisplayed)
                addstr_wrap(stdscr, 3+i, len(ndisplayed), " " * (width - len(ndisplayed)))
            stdscr.attroff(curses.color_pair(7))
            stdscr.attroff(curses.A_BOLD)

        # Stats section
        if mode_str == 'Stats':
            if cur_time - last_stats > REFRESH:
                last_stats = cur_time
                stats = get_stats(regex_str)
            #stdscr.addstr(1, 0, stats_title)
            #stdscr.addstr(1, len(stats_title), " " * (width - len(stats_title)))
            i = 0
            color_pair_i = 4
            for output in stats:
                stdscr.attron(curses.color_pair(color_pair_i))
                stdscr.attron(curses.A_BOLD)
                for s in output:
                    i += 1
                    if i+4 > height:
                        break
                    addstr_wrap(stdscr, 1+i, 0, s)
                    addstr_wrap(stdscr, 1+i, len(s), " " * (width - len(s)))
                i += 1
                stdscr.attroff(curses.color_pair(color_pair_i))
                stdscr.attroff(curses.A_BOLD)
                color_pair_i += 1
        if can_draw:
            stdscr.refresh()

        k = stdscr.getch()
        if k == -1:
            if not filter_capture:
                time.sleep(0.15)
            else:
                time.sleep(0.05)
        elif k == ord('f'):
            filter_capture = True
        elif not filter_capture and k == ord('o') and mode_str == 'Stats':
            mode_str = 'Orders'
        elif not filter_capture and k == ord('s') and mode_str == 'Orders':
            mode_str = 'Stats'
        elif not filter_capture and k == ord('q'):
            break


def main():
    while 1:
        do_restart = False
        try:
            curses.wrapper(draw_menu)
        except KeyboardInterrupt:
            break
        except Exception as err:
            do_restart = True
            endwin()
            print('An unexpected error: {}'.format(err))
            print('Restarting...')
        if not do_restart:
            break
        time.sleep(3)

if __name__ == "__main__":
    main()



"""CoinbasePro exchange module"""
import time
import configparser
import typing as t
import cbpro
from .exceptions import ExchangeError, ExchangeGetOrdersError, ExchangeAuthError
from .exceptions import ExchangeBuyLimitError, ExchangeBuyMarketError, ExchangeSellLimitError
from .exceptions import ExchangeSellMarketError, ExchangeProductInfoError, ExchangeCancelError
from .exceptions import ExchangeFeesError, ExchangeWalletError
from .base import BaseExchange, ProductInfo, Decimal

def _api_response_check(response, exception_to_raise):
    """Raise exception_to_raise if API response contains 'message'. If response['message']
    exists, this is _always_ (I think) and error scenario with CoinbasePro
    """
    if 'message' in response:
        raise exception_to_raise(response['message'])

class CoinbasePro(BaseExchange):
    """CoinbasePro exchange"""
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=no-member
    def __init__(self, config: configparser.ConfigParser) -> None:
        self.usd_decimal_places = 2
        self.size_decimal_places = 8
        super().__init__(config)
        self._last_call = time.time()

    def _rate_limit(self):
        cur_time = time.time()
        if cur_time - self._last_call < 2:
            sleep_time = cur_time - self._last_call
            if sleep_time > 0:
                time.sleep(sleep_time)
            else:
                time.sleep(2.1)
        self._last_call = time.time()

    def _wrap_client(self, method: str, *args, **kwargs):
        self._rate_limit()
        meth = getattr(self.client, method, None)
        for error in range(12):
            try:
                return meth(*args, **kwargs)
            except Exception as err:
                print('WARNING: exchange client error retry: {}'.format(err))
                errors += 1
                if error > 9:
                    raise
                time.sleep(10)

    def authenticate(self) -> cbpro.AuthenticatedClient:
        key = self.config['exchange'].get('key')
        passphrase = self.config['exchange'].get('passphrase')
        b64secret = self.config['exchange'].get('b64secret')
        self.client = cbpro.AuthenticatedClient(key, b64secret, passphrase)
        test = self.client.get_accounts()
        _api_response_check(test, ExchangeAuthError)
        return self.client

    def get_price(self) -> Decimal:
        ticker = self._wrap_client('get_product_ticker', product_id=self.coin)
        _api_response_check(ticker, ExchangeError)
        price = Decimal(ticker['price'])
        return price

    def get_precisions(self) -> ProductInfo:
        product_info = self.get_product_info()
        assert product_info is not None, 'Product info must be set.'
        # Set how many decimal places/precision price and size can have
        base_increment = '%.12f' % (product_info.base_increment)
        quote_increment = '%.12f' % (product_info.quote_increment)
        self.size_decimal_places = base_increment.split('1')[0].count('0')
        self.usd_decimal_places = quote_increment.split('1')[0].count('0')
        return (self.size_decimal_places, self.usd_decimal_places)

    def get_product_info(self) -> ProductInfo:
        product_info = None
        products = self._wrap_client('get_products')
        _api_response_check(products, ExchangeProductInfoError)
        for product in products:
            if product['id'] == self.coin:
                product_info = ProductInfo(product)
                product_info.digest()
                break
        assert product_info is not None, 'Product info must be set.'
        return product_info

    def get_usd_wallet(self) -> Decimal:
        accounts = self._wrap_client('get_accounts')
        _api_response_check(accounts, ExchangeWalletError)
        for account in accounts:
            if account['currency'] == 'USD':
                wallet = Decimal(account['available'])
                break
        assert wallet is not None, 'USD wallet was not found.'
        return wallet

    def get_open_sells(self) -> t.List[t.Mapping[str, Decimal]]:
        orders = self._wrap_client('get_orders')
        _api_response_check(orders, ExchangeGetOrdersError)
        open_sells = []
        for order in orders:
            if order['side'] == 'sell' and order['product_id'] == self.coin:
                order['price'] = Decimal(order['price'])
                order['size'] = Decimal(order['size'])
                open_sells.append(order)
        return open_sells

    def get_fees(self) -> t.Tuple[Decimal, Decimal, Decimal]:
        """pypi cbpro version doesn't have my get_fees() patch, so manually query it"""
        # pylint: disable=protected-access
        #fees = self.client._send_message('get', '/fees')
        fees = self._wrap_client('_send_message', 'get', '/fees')
        _api_response_check(fees, ExchangeFeesError)
        maker_fee = Decimal(fees['maker_fee_rate'])
        taker_fee = Decimal(fees['taker_fee_rate'])
        usd_volume = Decimal(fees['usd_volume'])
        return (maker_fee, taker_fee, usd_volume)

    def buy_limit(self, price: Decimal, size: Decimal) -> dict:
        fixed_price = str(round(Decimal(price), self.usd_decimal_places))
        fixed_size = str(round(Decimal(size), self.size_decimal_places))
        response = self._wrap_client('place_limit_order',
            product_id=self.coin,
            side='buy',
            size=fixed_size,
            price=fixed_price,
        )
        _api_response_check(response, ExchangeBuyLimitError)
        return response

    def buy_market(self, funds: Decimal) -> dict:
        funds = str(round(Decimal(funds), self.usd_decimal_places))
        self.logit('buy_market: funds:{}'.format(funds))
        response = self._wrap_client('place_market_order',
            product_id=self.coin,
            side='buy',
            funds=funds,
        )
        _api_response_check(response, ExchangeBuyMarketError)
        return response

    def sell_limit(self, price: Decimal, size: Decimal) -> dict:
        fixed_price = str(round(Decimal(price), self.usd_decimal_places))
        fixed_size = str(round(Decimal(size), self.size_decimal_places))
        self.logit('sell_limit: price:{} size:{}'.format(fixed_price, fixed_size))
        response = self._wrap_client('place_limit_order',
            product_id=self.coin,
            side='sell',
            price=fixed_price,
            size=fixed_size,
        )
        _api_response_check(response, ExchangeSellLimitError)
        return response

    def sell_market(self, size: Decimal) -> dict:
        fixed_size = str(round(Decimal(size), self.size_decimal_places))
        self.logit('sell_market: size:{}'.format(fixed_size))
        response = self._wrap_client('place_market_order',
            product_id=self.coin,
            side='sell',
            size=fixed_size,
        )
        _api_response_check(response, ExchangeSellMarketError)
        return response

    def cancel(self, order_id: str) -> bool:
        response = self._wrap_client('cancel_order', order_id)
        _api_response_check(response, ExchangeCancelError)
        return response

    def get_order(self, order_id: str) -> dict:
        response = self._wrap_client('get_order', order_id)
        _api_response_check(response, ExchangeGetOrdersError)
        return response

    def get_hold_value(self) -> Decimal:
        return Decimal(-1)



from distutils.core import setup
import setuptools
from setuptools import find_packages

setup(
    name='Botic',
    version='1.1.3',
    author='Matth Ingersoll',
    author_email='matth@mtingers.com',
    packages=find_packages(),
    license='GPLv3',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/mtingers/botic',
    install_requires=[
        'filelock>=3.0.12',
        'cbpro>=1.1.4',
    ],
    entry_points={
        'console_scripts': [
            'botic=botic.cli:main',
            'boticp=botic.cli:main_persist',
            'botictop=botic.top:main',
            'boticdump=botic.dumpdata:main',
        ],
    },
    package_data={'botic': ['data/historical-btc.csv.gz']},
    include_package_data=True,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)





"""A place to add configuration defaults"""

# Available config options with defaults
CONFIG_DEFAULTS = {
    'exchange': [
        ('exchange_module', str, 'CoinbasePro'),
        ('key', str, ''),
        ('passphrase', str, ''),
        ('b64secret', str, ''),
    ],
    'general': [
        ('coin', str, 'BTC-USD'),
        ('sleep_seconds', float, 60),
        ('log_file', str, 'botic-btc.log'),
        ('data_file', str, 'botic-btc.data'),
        ('pause_file', str, 'bot.pause'),
        ('log_disabled', bool, False),
    ],
    'trader': [
        ('trader_module', str, 'Simple'),
        # Optional config for the trader module goes here.
        # All trader modules can have their own custom names for these key/value pairs.
        # The default config reader will get/set all values in the [trader] section as strings.
        # The trader_module is responsible for converting values as needed.
        # Examples:
        #('max_outstanding_sells', int, 10),
        #('max_buys_per_hour', int, 10),
        #('sell_target', Decimal, Decimal('1.25')),
        #('buy_barrier', Decimal, Decimal('0.5')),
        #('buy_percent', Decimal, Decimal('2.0')),
        #('buy_max', Decimal, Decimal('150.00')),
        #('buy_min', Decimal, Decimal('35.00')),
        #('stoploss_enable', bool, False),
        #('stoploss_percent', Decimal, Decimal('-7.0')),
        #('stoploss_seconds', int, 86400),
        #('stoploss_strategy', str, 'report'),
    ],
    'notify': [
        ('notify_only_sold', bool, False),
        ('mail_host', str, ''),
        ('mail_from', str, ''),
        ('mail_to', str, ''),
    ],
    'debug': [
        ('debug_response', bool, False),
        ('debug_log', str, 'botic-debug.log'),
    ],
}



"""Template and base class for adding exchange modules"""
import time
from decimal import Decimal
import typing as t
from abc import abstractmethod
import configparser
from ..basebot import BaseBot

class ProductInfo: # pylint: disable=too-few-public-methods
    """Crypto product information class that stores some important information for making buy/sell
    calculations.

    Args:
        config_path (str): The path to the configuration file

    Attributes:
        id (str): Crypto name and fiat currency
        display_name (str): Friendly crypto name
        base_currency (str): crypto name
        quote_currency (str): Fiat currency
        base_increment (Decimal):  Specifies the minimum increment for the base_currency.
        quote_increment (Decimal): Specifies the min order price as well as the price increment.
        base_min_size (Decimal): Min order size
        base_max_size (Decimal): Max order size
        min_market_funds (int): Min funds allowed in a market order
        max_market_funds (int): Max funds allowed in a market order
        status (str): Short description of the exchange status for this coin
        status_message (str): provides any extra information regarding the status if available.
        cancel_only (bool): indicates whether this product only accepts cancel requests for orders
        post_only (bool): indicates whether only maker orders can be placed. No orders will be
            matched when post_only mode is active.
        limit_only (bool): indicates whether this product only accepts limit orders.
        trading_disabled (bool): indicates whether trading is currently restricted on this product,
            this includes whether both new orders and order cancelations are restricted.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, product_info: dict) -> None:
        self.product_info = product_info
        self.config = {
            'id':'',
            'display_name':'',
            'base_currency':'',
            'quote_currency':'',
            'base_increment':Decimal('-1'),
            'quote_increment':Decimal('-1'),
            'base_min_size':Decimal('-1'),
            'base_max_size':Decimal('-1'),
            'min_market_funds':Decimal(-1),
            'max_market_funds':Decimal(-1),
            'status':'',
            'status_message':'',
            'cancel_only':False,
            'limit_only':False,
            'post_only':False,
            'trading_disabled':False,
            'margin_enabled':False,
        }
        self.display_name = ''
        self.base_currency = ''
        self.quote_currency = ''
        self.base_increment = Decimal('-1')
        self.quote_increment = Decimal('-1')
        self.base_min_size = Decimal('-1')
        self.base_max_size = Decimal('-1')
        self.min_market_funds = Decimal(-1)
        self.max_market_funds = Decimal(-1)
        self.status = ''
        self.status_message = ''
        self.cancel_only = False
        self.limit_only = False
        self.post_only = False
        self.trading_disabled = False
        self.margin_enabled = False
        self.digest()

    def	digest(self) -> None:
        """This can be overridden for handling different exchanges
        To do so, map the exchange product info to the self.config values as closely as possible.
        """
        for key,val in self.product_info.items():
            cast = type(self.config[key])
            setattr(self, key, cast(val))

class BaseExchange(BaseBot):
    """Base class of abstractmethods to implement for each exchange. It is important to note that
    each method sets AND returns the same value (e.g. self.price, return self.price).

    Args:
        config_path (str): The path to the configuration file

    Attributes:
        config (configparser.ConfigParser): ConfigParser object from file specified by config_path
    """
    def __init__(self, config: configparser.ConfigParser) -> None:
        super().__init__(config)
        self.client = None

    @abstractmethod
    def authenticate(self):
        """Authenticate/connect to the exchange using credentials from the config

        Sets:
            self.client

        Returns:
            any: Authenticated client object

        Raises:
            ExchangeAuthError
        """

    @abstractmethod
    def get_price(self) -> Decimal:
        """Get latest price of coin from exchange

        Returns:
            Decimal: Decimal value of API response

        Raises:
            PriceError
            ExchangeError

        Notes:
            Example API: https://docs.pro.coinbase.com/#get-product-ticker
        """

    @abstractmethod
    def get_product_info(self) -> ProductInfo:
        """Build ProductInfo object from API response and set price and size precision (how many
        decimal places each can have).

        Returns:
            ProductInfo: Returns the data as a ProductInfo object

        Raises:
            ExchangeProductInfoError

        Notes:
            Example API: https://docs.pro.coinbase.com/#products
        """

    @abstractmethod
    def get_usd_wallet(self) -> Decimal:
        """Get the value of USD wallet

        Returns:
            Decimal: Decimal value of USD wallet

        Raises:
            ExchangeWalletError

        Notes:
            Example API: https://docs.pro.coinbase.com/#accounts

        """

    @abstractmethod
    def get_open_sells(self) -> t.List[t.Mapping[str, Decimal]]:
        """Query exchange for an active list of open sell orders.

        Returns:
            list[dict]: Return a list of orders with values converted to Decimal where applicable

        Raises:
            ExchangeError

        Notes:
            Example API: https://docs.pro.coinbase.com/#orders
        """

    @abstractmethod
    def get_fees(self) -> t.Tuple[Decimal, Decimal]:
        """Get current maker and taker fees and optionally USD volume

        Returns:
            tuple: (maker_fee, taker_fee, usd_volume)

        Raises:
            ExchangeFeesError

        Notes:
            Example API: https://docs.pro.coinbase.com/#fees
        """

    @abstractmethod
    def buy_limit(self, price: Decimal, size: Decimal) -> dict:
        """Place a buy limit order

        Args:
            price (Decimal): Price per crypto
            size: (Decimal): Amount of base currency to buy

        Returns:
            dict: API response on success

        Raises:
            ExchangeBuyLimitError

        Notes:
            Example API: https://docs.pro.coinbase.com/#orders
        """

    @abstractmethod
    def buy_market(self, funds: Decimal) -> dict:
        """Place a buy market order

        Args:
            funds (Decimal): How much of your wallet to use

        Returns:
            dict: API response on success

        Raises:
            ExchangeBuyMarketError

        Notes:
            Example API: https://docs.pro.coinbase.com/#orders
        """

    @abstractmethod
    def sell_limit(self, price: Decimal, size: Decimal) -> dict:
        """Place a sell limit order

        Args:
            price (Decimal): Price per crypto
            size: (Decimal): Amount of base currency to sell

        Returns:
            dict: API response on success

        Raises:
            ExchangeSellError

        Notes:
            Example API: https://docs.pro.coinbase.com/#orders
        """

    @abstractmethod
    def sell_market(self, size: Decimal) -> dict:
        """Place a sell market order

        Args:
            size: (Decimal): Amount of base currency to sell

        Returns:
            dict: API response on success

        Raises:
            ExchangeSellError

        Notes:
            Example API: https://docs.pro.coinbase.com/#orders
        """

    @abstractmethod
    def cancel(self, order_id: str) -> bool:
        """Cancel an order by it's ID

        Args:
            order_id (str): The order ID to cancel

        Returns:
            bool: True of successful, else False

        Raises:
            ExchangeCancelError

        Notes:
            Example API: https://docs.pro.coinbase.com/#cancel-an-order
        """

    @abstractmethod
    def get_order(self, order_id: str) -> dict:
        """Get order by ID

        Args:
            order_id (str): The order ID to get

        Returns:
            dict: API response as a dictionary

        Notes:
            Example API: https://docs.pro.coinbase.com/#get-an-order
        """

    @abstractmethod
    def get_precisions(self) -> tuple:
        """Get usd and size precisions/decimal places

        Returns:
            tuple: (size_decimal_places, usd_decimal_places)
        """

    @abstractmethod
    def get_hold_value(self) -> Decimal:
        """Get value of outstanding sell orders without fees

        Returns:
            Decimal: total value of outstanding sell orders at current price
        """

    def get_time(self) -> float:
        """Optional override: Return the time based off of what the exchange sees.
        This can be useful for running backtesting and emulating timing in the past.

        Returns:
            float: epoch timestamp
        """
        # pylint: disable=no-self-use
        return time.time()
